from typing import List, Optional, Tuple

import matplotlib as mpl
from matplotlib.axes._axes import Axes
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from nannernest.nesting import BoundingBox
from nannernest.segmentation import Segment
from nannernest.slicing import BananaCircle
from nannernest.slices import Slice

plt.rcParams["image.origin"] = "lower"


def mask_to_color(
    mask: np.ndarray,
    color: List[float],
    thresh: Optional[float] = None,
    alpha: float = 0.5,
) -> np.ndarray:
    rows, cols = mask.shape
    out = np.ones((rows, cols, 4), dtype=np.float32)
    out[..., :3] = color
    if thresh:
        threshed = mask > thresh
        out[threshed, 3] *= alpha

        out[~threshed, 3] *= 0
    else:
        out[..., 3] *= mask * alpha
    return out


def plot_segment(
    ax: Axes, segment: Segment, mask_color: List[float], fig_scaler: float
) -> Axes:
    ax.add_patch(
        plt.Rectangle(
            (segment.box[0], segment.box[1]),
            segment.box[2] - segment.box[0],
            segment.box[3] - segment.box[1],
            fill=False,
            edgecolor="red",
            linewidth=3.5 * fig_scaler,
        )
    )
    ax.text(
        segment.box[0],
        segment.box[1] - 2,
        "{:s} {:.3f}".format(segment.name, segment.score),
        bbox=dict(facecolor="blue", alpha=0.5),
        fontsize=14 * fig_scaler,
        color="white",
    )
    ax.imshow(mask_to_color(segment.mask, mask_color))
    return ax


def plot_slices(ax: Axes, slices: List[Slice], fig_scaler: float) -> Axes:
    cmap = sns.color_palette("Paired")

    nested_slices = [s for s in slices if s.nested_coords is not None]
    if nested_slices:
        # Copy nested slices over and only plot these slices.
        slices = list(nested_slices)
    # else: Do nothing. We will plot all possible slices on the banana.

    for slice_ in slices:
        point_array = np.array(
            list(
                zip(
                    slice_.upper_right,
                    slice_.lower_right,
                    slice_.lower_left,
                    slice_.upper_left,
                )
            )
        ).T
        ax.add_patch(
            mpl.patches.Polygon(
                point_array,
                closed=True,
                alpha=0.5,
                edgecolor="black",
                facecolor=cmap[slice_.index % len(cmap)],
            )
        )
        ax.annotate(
            f"{slice_.index}",
            slice_.centroid,
            ha="center",
            va="center",
            fontsize=12 * fig_scaler,
        )

        if slice_.nested_coords is not None:
            ax.add_patch(
                mpl.patches.Polygon(
                    slice_.nested_coords,
                    closed=True,
                    alpha=0.5,
                    edgecolor="black",
                    facecolor=cmap[slice_.index % len(cmap)],
                )
            )
            centroid = slice_.nested_coords.mean(axis=0)
            ax.annotate(
                f"{slice_.index}",
                centroid,
                ha="center",
                va="center",
                fontsize=12 * fig_scaler,
            )

    return ax


def plot(
    image: np.ndarray,
    banana: Optional[Segment] = None,
    bread: Optional[Segment] = None,
    slices: Optional[List[Slice]] = None,
    banana_skeleton: Optional[Tuple[np.ndarray, np.ndarray]] = None,
    banana_circle: Optional[BananaCircle] = None,
    banana_centroid: Optional[Tuple[float, float]] = None,
    bread_box: Optional[BoundingBox] = None,
    mask_color: Optional[List[float]] = None,
    output: Optional[str] = "perfect_sandwich.jpg",
    show: bool = False,
    dpi: int = 80,
) -> Optional[Axes]:

    # Default color is green   [  R,   G,   B]
    mask_color = mask_color or [0.0, 1.0, 0.0]

    # The figure scaling stuff is from https://stackoverflow.com/a/34769840
    height, width, _ = np.array(image).shape

    # What size does the figure need to be in inches to fit the image?
    figsize = width / float(dpi), height / float(dpi)
    # Use fig_scaler to scale all fonts relative to the figure size.
    fig_scaler = max(figsize) / 10

    # Create a figure of the right size with one axes that takes up the full figure
    fig = plt.figure(figsize=figsize)
    ax = fig.add_axes([0, 0, 1, 1])
    ax.axis("off")

    ax.imshow(image, aspect="equal")

    for segment in (banana, bread):
        if segment is not None:
            ax = plot_segment(ax, segment, mask_color, fig_scaler)

    if banana_skeleton is not None:
        ax.plot(
            *banana_skeleton, ".", linewidth=1 * fig_scaler, markersize=10 * fig_scaler
        )

    if banana_circle is not None:
        ax.plot(
            *banana_circle.draw(), "orange", linestyle="--", linewidth=2 * fig_scaler,
        )
        # Center of the circle
        ax.plot(
            *banana_circle.center,
            "o",
            markerfacecolor="orange",
            markeredgecolor="None",
            markersize=15 * fig_scaler,
        )

    if banana_centroid is not None:
        # Mask centroid
        ax.plot(
            banana_centroid[1], banana_centroid[0], "or", markersize=15 * fig_scaler,
        )

    if bread_box is not None:
        num_points = len(bread_box.vertices)
        for i in range(num_points):
            ax.plot(
                [bread_box.vertices[i, 0], bread_box.vertices[(i + 1) % num_points, 0]],
                [bread_box.vertices[i, 1], bread_box.vertices[(i + 1) % num_points, 1]],
                color="black",
                linestyle="dashed",
                linewidth=2 * fig_scaler,
            )

    if slices is not None:
        ax = plot_slices(ax, slices, fig_scaler)

    ax.set(xlim=[-0.5, width - 0.5], ylim=[-0.5, height - 0.5], aspect=1)

    if output is not None:
        fig.savefig(output, dpi=dpi, transparent=True)

    if show:
        plt.show()
    else:
        return ax



from dataclasses import dataclass, field
from glob import glob
import os
from typing import List, Tuple
import xml.etree.ElementTree as etree

import nest2D
import numpy as np
from scipy.spatial import ConvexHull
from skimage import measure
from svgpath2mpl import parse_path
import sympy
from sympy.solvers import solve
from sympy import Symbol


from nannernest.segmentation import Segment
from nannernest.slices import Slice


_MM_IN_COORD_UNITS = 1000000
# L is a magic number from nest2D and is related to the fixed size with which the
# SVG output is written.
_NEST_L = 150000000


@dataclass
class BoundingBox:
    vertices: np.ndarray
    area: float
    height: float
    width: float
    angle: float
    reference: np.ndarray = field(init=False)

    def __post_init__(self):
        self.reference = self.vertices[1, :]


def calc_elliptical_polygon_scalers(num_points: int) -> Tuple[float, float]:
    theta = 2 * np.pi / (num_points - 1)
    a, b = Symbol("a", positive=True), Symbol("b", positive=True)
    da, db = Symbol("da", positive=True), Symbol("db", positive=True)
    # fmt: off
    f1 = (
        (
            (-b * sympy.sin(theta/2))
            / ((a + da) - a * sympy.cos(theta/2))
        )
        - (
            ((b * sympy.sin(theta/2) - (b + db) * sympy.sin(theta))
             / (a * sympy.cos(theta/2) - (a + da) * sympy.cos(theta)))
        )
    )

    f2 = (
        (
            ((b + db) * sympy.sin(theta) - b * sympy.sin(3 * theta / 2))
            / ((a + da) * sympy.cos(theta) - a * sympy.cos(3 * theta / 2))
        )
        - (
            ((b * sympy.sin(3 * theta / 2) - (b + db) * sympy.sin(2 * theta))
             / ((a * sympy.cos(3 * theta / 2) - (a + da) * sympy.cos(2 * theta))))
        )
    )
    # fmt: on
    res = solve([f1, f2], da, db, dict=True)

    # Below tells us what to scale a and b by in our
    # ellipse equation.
    # The original derivation was (a + da)
    # The solutions are in the form of da = coeff * a
    # So, we return in the form of (1 + coeff) * a
    major_scaler = 1 + res[0][da].coeff(a)
    minor_scaler = 1 + res[0][db].coeff(b)
    return major_scaler, minor_scaler


def ellipse_to_polygon(
    slice_: Slice, num_points: int, major_scaler: float, minor_scaler: float
) -> List[Tuple[float, float]]:
    theta = np.linspace(0, 2 * np.pi, num_points + 1)
    # Divide major axis by 2 so that it turns into a radius
    x = (major_scaler * slice_.major_axis / 2) * np.cos(theta)
    y = (minor_scaler * slice_.minor_axis / 2) * np.sin(theta)
    # Reverse the order so that it goes clockwise because nest2D assigns positive
    # (negative) area to clockwise (counterclockwise) polygons
    return list(zip(x, y))[::-1]


def polygons_to_nest_items(polygons: List[List[Tuple[float, float]]], scaler: float):
    items = []
    for polygon in polygons:
        items.append(
            nest2D.Item(
                [nest2D.Point(int(x * scaler), int(y * scaler)) for (x, y) in polygon]
            )
        )
    return items


def max_slices(items, box):
    best = None
    for idx in range(1, len(items)):
        pgrp = nest2D.nest(items[:idx], box)
        if len(pgrp) > 1:
            break
        else:
            best = idx

    if best is None:
        raise RuntimeError("Cannot find good nesting")
    return best


def nest(polygons, bread_box):
    L_x = int(bread_box.width)
    L_y = int(bread_box.height)
    # L_x = int(bread.box[2] - bread.box[0])
    # L_y = int(bread.box[3] - bread.box[1])
    svg_scaler = _NEST_L / L_x
    items = polygons_to_nest_items(polygons, svg_scaler)
    box = nest2D.Box(_NEST_L, int(_NEST_L * L_y / L_x))
    max_slice_idx = max_slices(items, box)
    pgrp = nest2D.nest(items[:max_slice_idx], box)
    writer = nest2D.SVGWriter()
    writer.write_packgroup(pgrp)
    writer.save()


def assign_nested_coords(
    slices: List[Slice], svg_file: str, bread_box: BoundingBox
) -> List[Slice]:
    Y_MAX = 210
    # L_x = int(bread.box[2] - bread.box[0])
    # L_y = int(bread.box[3] - bread.box[1])
    L_x = int(bread_box.width)
    L_y = int(bread_box.height)
    L_y_mm = _NEST_L * L_y / L_x / _MM_IN_COORD_UNITS
    reverse_scaler = _MM_IN_COORD_UNITS * L_x / _NEST_L

    tree = etree.parse(svg_file)
    root = tree.getroot()
    path_elems = root.findall(".//{http://www.w3.org/2000/svg}path")
    paths = [parse_path(elem.attrib["d"]) for elem in path_elems]

    # Rotation matrix to rotate to bread box angle
    rotation_mat = np.array(
        [
            [np.cos(bread_box.angle), -np.sin(bread_box.angle)],
            [np.sin(bread_box.angle), np.cos(bread_box.angle)],
        ]
    )

    for idx, (slice_, path) in enumerate(zip(slices, paths)):

        # Translate to origin
        vertices = np.array(path.vertices)
        vertices[:, 1] -= Y_MAX - L_y_mm
        # Scale to original image size
        scaled = vertices * reverse_scaler

        # Rotate polygons so that they're back at the bread box angle
        # dot product is 2 x 2 * 2 x num_points -> 2 x num_points
        # Transpose that, and we're back to num_points x 2
        scaled = np.dot(rotation_mat, scaled.T).T

        # Translate to bread
        # scaled[:, 0] += bread.box[0]
        # scaled[:, 1] += bread.box[1]
        scaled[:, 0] += bread_box.reference[0]
        scaled[:, 1] += bread_box.reference[1]

        slice_.nested_coords = scaled

    return slices


def minimum_bounding_box(hull_points_2d: np.ndarray) -> BoundingBox:
    """
    Adapted from this blog post
    https://chadrick-kwag.net/python-implementation-of-rotating-caliper-algorithm/

    hull_points_2d: array of hull points. each element should have [x,y] format
    """

    # Compute edges (x2-x1,y2-y1)
    edges = np.zeros((len(hull_points_2d) - 1, 2))  # empty 2 column array
    for i in range(len(edges)):
        edge_x = hull_points_2d[i + 1, 0] - hull_points_2d[i, 0]
        edge_y = hull_points_2d[i + 1, 1] - hull_points_2d[i, 1]
        edges[i] = [edge_x, edge_y]

    # Calculate edge angles   atan2(y/x)
    edge_angles = np.zeros((len(edges)))  # empty 1 column array
    for i in range(len(edge_angles)):
        edge_angles[i] = np.arctan2(edges[i, 1], edges[i, 0])

    # Check for angles in 1st quadrant
    for i in range(len(edge_angles)):
        # want strictly positive answers
        edge_angles[i] = np.abs(edge_angles[i] % (np.pi / 2))

    # Remove duplicate angles
    edge_angles = np.unique(edge_angles)

    min_bbox = None
    min_area = np.finfo("float").max

    for i in range(len(edge_angles)):

        # Create rotation matrix to shift points to baseline
        # R = [ cos(theta)      , cos(theta-PI/2)
        #       cos(theta+PI/2) , cos(theta)     ]
        R = np.array(
            [
                [np.cos(edge_angles[i]), np.cos(edge_angles[i] - (np.pi / 2))],
                [np.cos(edge_angles[i] + (np.pi / 2)), np.cos(edge_angles[i])],
            ]
        )

        # Apply this rotation to convex hull points
        rot_points = np.dot(R, np.transpose(hull_points_2d))  # 2x2 * 2xn

        # Find min/max x,y points
        min_x = np.nanmin(rot_points[0], axis=0)
        max_x = np.nanmax(rot_points[0], axis=0)
        min_y = np.nanmin(rot_points[1], axis=0)
        max_y = np.nanmax(rot_points[1], axis=0)

        # Calculate height/width/area of this bounding rectangle
        width = max_x - min_x
        height = max_y - min_y
        area = width * height

        if area > min_area:
            continue
        else:
            min_area = area

        # Calculate center point and restore to original coordinate system
        center_x = (min_x + max_x) / 2
        center_y = (min_y + max_y) / 2
        center_point = np.dot([center_x, center_y], R)

        # Calculate corner points and restore to original coordinate system
        corner_points = np.zeros((4, 2))  # empty 2 column array
        corner_points[0] = np.dot([max_x, min_y], R)
        corner_points[1] = np.dot([min_x, min_y], R)
        corner_points[2] = np.dot([min_x, max_y], R)
        corner_points[3] = np.dot([max_x, max_y], R)

        min_bbox = BoundingBox(
            vertices=corner_points,
            area=area,
            height=height,
            width=width,
            angle=edge_angles[i],
        )
    if min_bbox is None:
        raise RuntimeError("Unable to find a minimum bounding box")
    return min_bbox


def make_bread_convex_hull(bread: Segment, mask_threshold: float) -> np.ndarray:
    props = measure.regionprops((bread.mask > mask_threshold).astype(int), bread.mask)[
        0
    ]
    x_offset, y_offset, *_ = props.bbox
    convex_x, convex_y = np.where(props.convex_image)
    convex_x += x_offset
    convex_y += y_offset
    points = np.vstack((convex_x, convex_y)).T
    convex_hull = ConvexHull(points)
    convex_hull = points[convex_hull.vertices, ::-1]  # Flip it around
    return convex_hull


def make_bread_box(bread: Segment, mask_threshold: float) -> BoundingBox:
    bread_chull = make_bread_convex_hull(bread, mask_threshold)
    return minimum_bounding_box(bread_chull)


def run(
    slices: List[Slice],
    bread: Segment,
    mask_threshold: float = 0.6,
    num_points: int = 30,
) -> Tuple[List[Slice], BoundingBox]:
    major_scaler, minor_scaler = calc_elliptical_polygon_scalers(num_points)
    polygons = [
        ellipse_to_polygon(slice_, num_points, major_scaler, minor_scaler)
        for slice_ in slices
    ]
    bread_box = make_bread_box(bread, mask_threshold)
    nest(polygons, bread_box)

    SVG_FILE = "out.svg"
    slices = assign_nested_coords(slices, SVG_FILE, bread_box)

    # Delete any svg files that were made in the nesting process. If multiple bins are
    # required for nesting, then multiple svg files numbered out1.svg, out2.svg, etc...
    # will be created.
    for f in glob("out*.svg"):
        os.remove(f)
    return slices, bread_box



from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Tuple, Union
import warnings

import numpy as np
from PIL import Image, ImageOps
import torch
from torchvision.models.detection import maskrcnn_resnet50_fpn
from torchvision import transforms

from nannernest.exceptions import SegmentNotFound


SegmentationOutput = Dict[str, torch.Tensor]


@dataclass
class Segment:
    label: np.int64
    name: str = field(init=False)
    score: np.float32
    mask: np.ndarray
    box: np.ndarray

    def __post_init__(self):
        self.name = coco_category_name(self.label)
        self.mask = self.mask.squeeze()

    def __eq__(self, other):
        return (
            self.label == other.label
            and self.name == other.name
            and self.score == other.score
            and np.allclose(self.mask, other.mask)
            and np.allclose(self.box, other.box)
        )


def load_model():
    return maskrcnn_resnet50_fpn(pretrained=True)


def segment_image(
    image_path: Path, model: torch.nn.Module
) -> Tuple[np.ndarray, SegmentationOutput]:
    model = model.eval()

    preprocess = transforms.Compose([transforms.ToTensor()])

    input_image = Image.open(image_path)

    # Flip the image here so that rendering the
    # image with the origin of the image array
    # at the lower left (like a regular plot)
    # makes the image look rightside up.
    input_image = ImageOps.flip(input_image)

    input_tensor = preprocess(input_image)
    # Convert to a mini-batch
    input_batch = input_tensor.unsqueeze(0)

    if torch.cuda.is_available():
        input_batch = input_batch.to("cuda")
        model = model.to("cuda")

    with torch.no_grad():
        with warnings.catch_warnings():
            # There's a pytorch deprecation warning triggered inside of torchvision.
            # I'd like to not have to filter this warning at some point.
            warnings.filterwarnings("ignore", category=UserWarning)
            output = model(input_batch)

    return input_image, output[0]


def coco_category_name(label: int) -> str:
    COCO_INSTANCE_CATEGORY_NAMES = [
        "__background__",
        "person",
        "bicycle",
        "car",
        "motorcycle",
        "airplane",
        "bus",
        "train",
        "truck",
        "boat",
        "traffic light",
        "fire hydrant",
        "N/A",
        "stop sign",
        "parking meter",
        "bench",
        "bird",
        "cat",
        "dog",
        "horse",
        "sheep",
        "cow",
        "elephant",
        "bear",
        "zebra",
        "giraffe",
        "N/A",
        "backpack",
        "umbrella",
        "N/A",
        "N/A",
        "handbag",
        "tie",
        "suitcase",
        "frisbee",
        "skis",
        "snowboard",
        "sports ball",
        "kite",
        "baseball bat",
        "baseball glove",
        "skateboard",
        "surfboard",
        "tennis racket",
        "bottle",
        "N/A",
        "wine glass",
        "cup",
        "fork",
        "knife",
        "spoon",
        "bowl",
        "banana",
        "apple",
        "sandwich",
        "orange",
        "broccoli",
        "carrot",
        "hot dog",
        "pizza",
        "donut",
        "cake",
        "chair",
        "couch",
        "potted plant",
        "bed",
        "N/A",
        "dining table",
        "N/A",
        "N/A",
        "toilet",
        "N/A",
        "tv",
        "laptop",
        "mouse",
        "remote",
        "keyboard",
        "cell phone",
        "microwave",
        "oven",
        "toaster",
        "sink",
        "refrigerator",
        "N/A",
        "book",
        "clock",
        "vase",
        "scissors",
        "teddy bear",
        "hair drier",
        "toothbrush",
    ]
    try:
        return COCO_INSTANCE_CATEGORY_NAMES[label]
    except IndexError:
        raise ValueError(f"Label {label} not found")


def to_segments(segmentation_output: SegmentationOutput) -> List[Segment]:
    boxes = segmentation_output["boxes"].cpu().numpy()
    labels = segmentation_output["labels"].cpu().numpy()
    scores = segmentation_output["scores"].cpu().numpy()
    masks = segmentation_output["masks"].cpu().numpy()
    segments = []
    for box, label, score, mask in zip(boxes, labels, scores, masks):
        segments.append(Segment(label=label, score=score, mask=mask, box=box))
    return segments


def find_best_segment(
    segments: List[Segment], labels: Union[str, List[str]]
) -> Segment:
    if isinstance(labels, str):
        labels = [labels]

    best_segment = None
    for segment in segments:
        if segment.name in labels:
            if best_segment is None or segment.score > best_segment.score:
                best_segment = segment
    if best_segment is None:
        raise SegmentNotFound(f"None of these labels {labels} detected in image")
    return best_segment


def run(image_path: Path) -> Tuple[np.ndarray, Segment, Segment]:
    model = load_model()
    image, segmentation_output = segment_image(image_path, model)
    segments = to_segments(segmentation_output)
    banana = find_best_segment(segments, "banana")
    bread = find_best_segment(segments, ["sandwich", "cake"])

    return image, banana, bread



from typing import Tuple, Optional

import numpy as np


class Slice:
    def __init__(
        self,
        index: int,
        upper_right: Tuple[float, float],
        lower_right: Tuple[float, float],
        lower_left: Tuple[float, float],
        upper_left: Tuple[float, float],
        nested_coords: Optional[np.ndarray] = None,
        ratio: float = 0.85,
    ):
        self.index = index
        self.upper_right = upper_right
        self.lower_right = lower_right
        self.lower_left = lower_left
        self.upper_left = upper_left
        self.nested_coords = nested_coords
        self.ratio = ratio
        self.major_axis = self.calc_major_axis()

    @property
    def minor_axis(self) -> float:
        return self.major_axis * self.ratio

    @property
    def centroid(self) -> Tuple[float, float]:
        return tuple(
            np.array(
                [self.upper_right, self.lower_right, self.lower_left, self.upper_left]
            ).mean(axis=0)
        )

    def calc_major_axis(self) -> float:
        first = euclidean_distance(self.upper_right, self.lower_right)
        second = euclidean_distance(self.upper_left, self.lower_left)
        return (first + second) / 2


def euclidean_distance(pt1: Tuple[float, float], pt2: Tuple[float, float]) -> float:
    return np.sqrt((pt1[0] - pt2[0]) ** 2 + (pt1[1] - pt2[1]) ** 2)



from typing import List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import scipy.optimize
from skimage import measure
from skimage.morphology import skeletonize

from nannernest.slices import Slice

plt.rcParams["image.origin"] = "lower"

MASK_THRESH = 0.8


class BananaCircle:
    def __init__(
        self, center: Tuple[float, float], radius: float, num_points: int = 201
    ):
        self.theta = np.linspace(0, 2 * np.pi, num_points)
        self.center = center
        self.radius = radius

    @property
    def xc(self) -> float:
        return self.center[0]

    @property
    def yc(self) -> float:
        return self.center[1]

    def draw(self) -> Tuple[np.ndarray, np.ndarray]:
        return (
            self.xc + self.radius * np.cos(self.theta),
            self.yc + self.radius * np.sin(self.theta),
        )


def calc_banana_skeleton(
    mask: np.ndarray, threshold: float
) -> Tuple[np.ndarray, np.ndarray]:
    skeleton = skeletonize(mask > threshold)
    # coordinates of the barycenter
    skeleton_coords = np.where(skeleton)
    # Remember that the columns are the "x" direction
    # and the rows are the "y" direction
    # So switch them here
    skeleton_coords = (skeleton_coords[1], skeleton_coords[0])
    return skeleton_coords


def fit_banana_arc(
    banana_skeleton: Tuple[np.ndarray, np.ndarray]
) -> Tuple[Tuple[float, float], float]:
    """
    Fit a circle to the banana skeleton. The below is straight-up grabbed from
    https://scipy-cookbook.readthedocs.io/items/Least_Squares_Circle.html#Using-scipy.optimize.leastsq

    Parameters
    ----------
    banana_skeleton : tuple(np.ndarray, np.ndarray)
        The coordinates of the banana skeleton as a tuple of (x, y) where x and y are
        arrays.

    Returns
    -------
    center_est
        The estimated center of the circle
    r_est
        The estimated radius of the circle

    """

    def radial_distance(
        xc: float, yc: float, x: np.ndarray, y: np.ndarray
    ) -> np.ndarray:
        """calculate the distance of each 2D points from the center (xc, yc)"""
        return np.sqrt((x - xc) ** 2 + (y - yc) ** 2)

    def radial_distance_differences(
        circle: Tuple[float, float], x: np.ndarray, y: np.ndarray
    ) -> np.ndarray:
        xc, yc = circle
        """ 
        calculate the algebraic distance between the data points and the mean circle 
        centered at c=(xc, yc) 
        """
        r = radial_distance(xc, yc, x, y)
        return r - r.mean()

    skeleton_center = np.mean(banana_skeleton[0]), np.mean(banana_skeleton[1])

    center_est, ier = scipy.optimize.leastsq(
        radial_distance_differences, skeleton_center, args=banana_skeleton
    )

    r_dist = radial_distance(*center_est, *banana_skeleton)
    r_est = r_dist.mean()

    return center_est, r_est


def calc_banana_centroid(mask: np.ndarray, mask_thresh: float) -> Tuple[float, float]:
    """Given a segmentation banana mask, find the centroid of the masked region"""
    return measure.regionprops((mask > mask_thresh).astype(int), mask)[0].centroid


def circle(
    x: float, y: float, r: float, theta: Union[np.ndarray, float]
) -> Tuple[Union[float, np.ndarray], Union[np.ndarray, float]]:
    return (x + r * np.cos(theta), y + r * np.sin(theta))


def create_phi_space(
    banana_centroid: Tuple[float, float],
    banana_circle: BananaCircle,
    num_points: int = 201,
) -> np.ndarray:
    """
    Create an array of angles centered at the angle that the banana centroid makes with
    respect to the banana circle center (phi_0). The angles span a 3 pi range. For the
    life of me, I can't remember why the angles span a 3 pi range, but the results
    seem to be worse if I reduce the range.
    """

    # Get phi_0, the angle of the centroid wrt the circle center.
    # Specifically, treat the horizontal axis of the image as zero degrees.
    # We will measure all angles relative to the angle that the
    # line between circle center and centroid makes.
    dy = banana_centroid[0] - banana_circle.yc
    dx = banana_centroid[1] - banana_circle.xc
    phi_0 = np.arctan2(dy, dx)

    # For the life of me, I can't remember why the below spans a
    phi_space = np.linspace(phi_0 - 3 * np.pi / 4, phi_0 + 3 * np.pi / 4, num_points)
    return phi_space


def assemble_profiles(phi_space, mask, banana_circle, linewidth=5):

    profiles = []
    for phi in phi_space:
        profiles.append(
            measure.profile_line(
                mask.T,  # For some reason gotta take the transpose
                banana_circle.center,
                circle(
                    banana_circle.xc, banana_circle.yc, 3 * banana_circle.radius, phi
                ),
                linewidth=linewidth,
                mode="constant",
            )[: 2 * int(banana_circle.radius)]
        )

    profiles = np.vstack(profiles)
    return profiles


def find_edges(mat: np.ndarray, axis: int = 1) -> np.ndarray:
    return np.sort(np.argsort(-np.abs(np.diff(np.max(mat, axis=axis), append=0)))[:2])


def find_banana_termini(
    profiles: np.ndarray, mask_threshold: float, frac: float = 0.30
) -> Tuple[int, int]:
    """Find the banana start and end indices of `profiles`"""
    edges = find_edges(profiles > mask_threshold, axis=1)
    length = edges[1] - edges[0]
    delta = int(np.round(frac * length))

    edge1_area = profiles[edges[0] : edges[0] + delta, :].mean()
    edge2_area = profiles[edges[1] - delta : edges[1], :].mean()

    start, end = edges[np.argsort([edge1_area, edge2_area])]

    end_area = max([edge1_area, edge2_area])

    # Now, remove the stem
    # The sign handles the case where start is > end.
    sign = np.sign(end - start)
    sliding_start_areas = np.array(
        [
            profiles[start + sign * i : start + sign * (i + delta) : sign, :].mean()
            for i in np.arange(delta)
        ]
    )
    stem = np.argmin(np.abs(sliding_start_areas - end_area))
    return start + sign * stem, end


def slice_bottom_top(
    profile_slices: np.ndarray, threshold: float
) -> Tuple[float, float]:
    bottom, top = np.sort(
        np.argsort(np.diff((profile_slices > threshold).max(axis=0)))[-2:]
    )
    return bottom, top


def assemble_slices(
    num_slices: int,
    start: float,
    end: float,
    phi_space: np.ndarray,
    banana_circle: BananaCircle,
    profiles: np.ndarray,
    mask_threshold: float,
    ratio: float,
) -> List[Slice]:
    slice_points = np.round(np.linspace(start, end, num_slices + 1)).astype(np.int32)
    slices = []
    for index, (slice_start, slice_end) in enumerate(
        zip(slice_points[:-1], slice_points[1:])
    ):
        slice_ = slice(slice_start, slice_end, np.sign(slice_end - slice_start))
        slice_bottom, slice_top = slice_bottom_top(profiles[slice_, :], mask_threshold)

        # TODO: Convert the slice class into one that either stores both the
        # bounding box coordinates _and_ the polar coordinates. We will need the
        # radius (slice_top - slice_bottom) in order to create the elliptical
        # cross-sectional view of the slices.
        upper_right = circle(
            banana_circle.xc, banana_circle.yc, slice_top, phi_space[slice_end]
        )
        lower_right = circle(
            banana_circle.xc, banana_circle.yc, slice_bottom, phi_space[slice_end]
        )
        lower_left = circle(
            banana_circle.xc, banana_circle.yc, slice_bottom, phi_space[slice_start]
        )
        upper_left = circle(
            banana_circle.xc, banana_circle.yc, slice_top, phi_space[slice_start]
        )

        slices.append(
            Slice(index, upper_right, lower_right, lower_left, upper_left, ratio=ratio)
        )
    return slices


def angular_slice(
    mask: np.ndarray,
    banana_circle: BananaCircle,
    banana_centroid: Tuple[float, float],
    mask_threshold: float,
    ratio: float,
    num_slices: int,
) -> List[Slice]:
    phi_space = create_phi_space(banana_centroid, banana_circle, num_points=201)
    profiles = assemble_profiles(phi_space, mask, banana_circle, linewidth=2)
    start, end = find_banana_termini(profiles, mask_threshold)

    slices = assemble_slices(
        # We'll throw out the first  and last slice, so let's cut 2 extra.
        num_slices + 2,
        start,
        end,
        phi_space,
        banana_circle,
        profiles,
        mask_threshold,
        ratio,
    )
    # Throw out the first and last slices.
    return slices[1:-1]


def run(
    mask: np.ndarray,
    num_slices: int = 22,
    peel_scaler: float = 0.8,
    mask_threshold: float = 0.6,
    ellipse_ratio: float = 0.85,
) -> Tuple[
    List[Slice], BananaCircle, Tuple[float, float], Tuple[np.ndarray, np.ndarray]
]:
    # The first step of slicing involves turning the banana into a parametrized circle.

    # Reduce the banana to a single curved line, 1 pixel wide.
    # See https://scikit-image.org/docs/dev/auto_examples/edges/plot_skeleton.html
    banana_skeleton = calc_banana_skeleton(mask, mask_threshold)

    # Fit a circle to the skeleton
    banana_center, banana_radius = fit_banana_arc(banana_skeleton)
    banana_circle = BananaCircle(banana_center, banana_radius)

    banana_centroid = calc_banana_centroid(mask, mask_threshold)

    slices = angular_slice(
        mask, banana_circle, banana_centroid, mask_threshold, ellipse_ratio, num_slices,
    )

    for s in slices:
        s.major_axis *= peel_scaler
    return slices, banana_circle, banana_centroid, banana_skeleton



import click_spinner
from pathlib import Path
import typer

from nannernest import nesting, segmentation, slicing, viz
from nannernest.exceptions import SegmentNotFound


_BANANA = "\U0001F34C"
_BREAD = "\U0001F35E"
_FG = typer.colors.GREEN


app = typer.Typer()


@app.command()
def cli(
    image_path: Path = typer.Argument(
        ..., help="Image file which contains bread and banana"
    ),
    num_slices: int = typer.Option(
        22,
        help="Total number of slices to cut the banana into. This number defines the slice thickness.",
    ),
    mask_threshold: float = typer.Option(0.6, help="Threshold of segmentation mask."),
    peel_scaler: float = typer.Option(
        0.8,
        help="Fraction of slice that is assumed to belong to banana insides versus the peel.",
    ),
    ellipse_ratio: float = typer.Option(
        0.85, help="Assumed ratio of minor axis to major axis of banana slice ellipses"
    ),
    plot_segmentation: bool = typer.Option(
        False, help="Whether or not to plot the segmentation masks"
    ),
    plot_slicing: bool = typer.Option(
        False, help="Whether or not to plot the slicing circle and skeleton"
    ),
    output: str = typer.Option("perfect_sandwich.jpg", help="Name of file to output"),
    suppress: bool = typer.Option(
        False,
        help="Whether or not to suppress the output file from being automatically opened",
    ),
):

    typer.secho(f"Segmenting image looking for {_BANANA} and {_BREAD}", fg=_FG)

    try:
        with click_spinner.spinner():
            image, banana, bread = segmentation.run(image_path)
    except SegmentNotFound:
        typer.secho(
            "Oh no! Can't find the banana or bread in the image", fg=typer.colors.RED,
        )
        raise SegmentNotFound

    typer.secho(f"Slicing the {_BANANA}", fg=_FG)
    with click_spinner.spinner():
        slices, banana_circle, banana_centroid, banana_skeleton = slicing.run(
            banana.mask,
            num_slices=num_slices,
            mask_threshold=mask_threshold,
            peel_scaler=peel_scaler,
            ellipse_ratio=ellipse_ratio,
        )

    typer.secho("Optimizing slice coverage", fg=_FG)
    with click_spinner.spinner():
        slices, bread_box = nesting.run(slices, bread, mask_threshold=mask_threshold)

    viz.plot(
        image,
        slices=slices,
        banana=banana if plot_segmentation else None,
        bread=bread if plot_segmentation else None,
        banana_skeleton=banana_skeleton if plot_slicing else None,
        banana_circle=banana_circle if plot_slicing else None,
        banana_centroid=banana_centroid if plot_slicing else None,
        bread_box=bread_box if plot_slicing else None,
        output=output,
    )

    typer.secho(f"Nested nanner saved to {output}", fg=_FG)
    if not suppress:
        typer.launch(output)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nannernest']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=7.2.0,<8.0.0',
 'click-spinner>=0.1.10,<0.2.0',
 'colorama>=0.4.3,<0.5.0',
 'matplotlib>=3.3.0,<4.0.0',
 'nest2D==0.4.3',
 'numpy>=1.19.1,<2.0.0',
 'scikit-image>=0.17.2,<0.18.0',
 'scipy>=1.5.2,<2.0.0',
 'seaborn>=0.10.1,<0.11.0',
 'svgpath2mpl>=0.2.1,<0.3.0',
 'sympy>=1.6.1,<2.0.0',
 'torch>=1.6.0,<2.0.0',
 'torchvision>=0.7.0,<0.8.0',
 'typer>=0.3.1,<0.4.0']

entry_points = \
{'console_scripts': ['nannernest = nannernest.cli:app']}

setup_kwargs = {
    'name': 'nannernest',
    'version': '0.1.4',
    'description': 'Optimal peanut butter and banana sandwiches',
    'long_description': "# nannernest\n\n![Python package](https://github.com/EthanRosenthal/nannernest/workflows/Python%20package/badge.svg?branch=master)\n\nA small package for optimizing banana coverage on peanut butter and banana sandwiches. See [blog post](https://www.ethanrosenthal.com/2020/08/25/optimal-peanut-butter-and-banana-sandwiches/) for more info.\n\n\n![assets/perfect_sandwich.jpg](assets/perfect_sandwich.jpg)\n\n\n## Installation\n\n`nannernest` is generally pip installable. Due to some C dependencies with the nesting library that I use [nest2D](https://github.com/markfink/nest2D), along with an outstanding [PR](https://github.com/markfink/nest2D/pull/2), I would recommend the following way to install everything:\n\n First, make sure you have [boost](https://www.boost.org/) and [cmake](https://cmake.org/) installed. If you are on Linux, then you may have `cmake` installed, and you can install `boost` with \n \n ```commandline\nsudo apt-get install libboost-all-dev \n```\n \n I'm on a Mac, and I installed `cmake` with conda and `boost` with brew:\n \n ```commandline\nconda install cmake\nbrew install boost\n```\n\nNext, pip install my fork of `nest2D`:\n\n```commandline\npip install git+https://github.com/EthanRosenthal/nest2D.git@download-dependencies\n```\n\nFinally, pip install `nannernest`\n\n```commandline\npip install nannernest\n```\n\n## Usage\n\nTake a top-down picture that contains your banana and at least one slice of bread. Pass the image in via command line:\n\n```commandline\n$ nannernest my_image.jpg\n```\n\n### CLI Details\n\n```commandline\n$ nannernest --help\nUsage: nannernest [OPTIONS] IMAGE_PATH\n\nArguments:\n  IMAGE_PATH  Image file which contains bread and banana  [required]\n\nOptions:\n  --num-slices INTEGER            Total number of slices to cut the banana\n                                  into. This number defines the slice\n                                  thickness.  [default: 22]\n\n  --mask-threshold FLOAT          Threshold of segmentation mask.  [default:\n                                  0.6]\n\n  --peel-scaler FLOAT             Fraction of slice that is assumed to belong\n                                  to banana insides versus the peel.\n                                  [default: 0.8]\n\n  --ellipse-ratio FLOAT           Assumed ratio of minor axis to major axis of\n                                  banana slice ellipses  [default: 0.85]\n\n  --plot-segmentation / --no-plot-segmentation\n                                  Whether or not to plot the segmentation\n                                  masks  [default: False]\n\n  --plot-slicing / --no-plot-slicing\n                                  Whether or not to plot the slicing circle\n                                  and skeleton  [default: False]\n\n  --output TEXT                   Name of file to output  [default:\n                                  perfect_sandwich.jpg]\n```\n",
    'author': 'Ethan Rosenthal',
    'author_email': 'ethan@ethanrosenthal.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/EthanRosenthal/nannernest',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
